/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/content/NodeRevisionDescriptor.java,v 1.19 2001/09/04 17:37:56 juergen Exp $
 * $Revision: 1.19 $
 * $Date: 2001/09/04 17:37:56 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.slide.content;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Date;
import java.util.StringTokenizer;
import java.util.Locale;
import java.util.TimeZone;
import java.text.SimpleDateFormat;
import java.text.DateFormat;
import java.text.ParseException;
import org.apache.slide.common.*;
import org.apache.slide.util.Messages;

/**
 * Node Revision Descriptor class.
 *
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.19 $
 */
public final class NodeRevisionDescriptor implements Serializable, Cloneable {
    
    
    // -------------------------------------------------------------- Constants
    
    
    /**
     * Creation date.
     */
    public static final String CREATION_DATE = "creationdate";
    
    
    /**
     * Last modification date.
     */
    public static final String LAST_MODIFIED = "getlastmodified";
    
    
    /**
     * Name.
     */
    public static final String NAME = "displayname";
    
    
    /**
     * Type.
     */
    public static final String TYPE = "resourcetype";
    
    
    /**
     * Source.
     */
    public static final String SOURCE = "source";
    
    
    /**
     * MIME type of the content.
     */
    public static final String CONTENT_TYPE = "getcontenttype";
    
    
    /**
     * Content language.
     */
    public static final String CONTENT_LANGUAGE = "getcontentlanguage";
    
    
    /**
     * Content length.
     */
    public static final String CONTENT_LENGTH = "getcontentlength";
    
    
    /**
     * ETag.
     */
    public static final String ETAG = "getetag";
    
    
    /**
     * Collection type.
     */
    public static final String COLLECTION_TYPE = "<collection/>";
    
    
    /**
     * HTTP date format.
     */
    protected static final SimpleDateFormat format =
        new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz", Locale.US);
    
    
    /**
     * Date formats using for Date parsing.
     */
    protected static final SimpleDateFormat formats[] = {
        new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz", Locale.US),
            new SimpleDateFormat("EEE MMM dd HH:mm:ss zzz yyyy", Locale.US),
            new SimpleDateFormat("EEEEEE, dd-MMM-yy HH:mm:ss zzz", Locale.US),
            new SimpleDateFormat("EEE MMMM d HH:mm:ss yyyy", Locale.US),
            new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'")
    };
    
    
    /**
     * Simple date format for the creation date ISO representation (partial).
     */
    protected static final SimpleDateFormat creationDateFormat =
        new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
    
    
    static {
        format.setTimeZone(TimeZone.getTimeZone("GMT"));
        creationDateFormat.setTimeZone(TimeZone.getTimeZone("GMT"));
        for (int i = 0; i < formats.length; i++) {
            formats[i].setTimeZone(TimeZone.getTimeZone("GMT"));
        }
    }
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Constructor.
     */
    public NodeRevisionDescriptor() {
        this.properties = new Hashtable();
        this.labels = new Vector();
        this.branchName = NodeRevisionDescriptors.MAIN_BRANCH;
        initDefaultProperties();
    }
    
    
    /**
     * Constructor for client applications.
     */
    public NodeRevisionDescriptor(long contentLength) {
        this();
        setContentLength(contentLength);
    }
    
    
    /**
     * Constructor used for retrieval. Can be used by client apps, but it is
     * mostly useless as number and creationDate fields are ignored during
     * revision creation.
     */
    public NodeRevisionDescriptor(NodeRevisionNumber number, String branchName,
                                  Vector labels, Hashtable properties) {
        this();
        this.number = number;
        if (branchName != null) {
            this.branchName = branchName;
        }
        this.labels = labels;
        setProperties(properties);
    }
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Branch name.
     */
    private String branchName;
    
    
    /**
     * Revision number.
     */
    private NodeRevisionNumber number;
    
    
    /**
     * Revision labels.
     */
    private Vector labels;
    
    
    /**
     * Properties field.
     */
    private Hashtable properties;
    
    
    // ------------------------------------------------------------- Properties
    
    
    /**
     * Branch name accessor.
     */
    public String getBranchName() {
        return this.branchName;
    }
    
    
    /**
     * Branch name mutator.
     */
    void setBranchName(String branchName) {
        this.branchName = branchName;
    }
    
    
    /**
     * Revision number accessor.
     */
    public NodeRevisionNumber getRevisionNumber() {
        return this.number;
    }
    
    
    /**
     * Revision number mutator.
     */
    void setRevisionNumber(NodeRevisionNumber number) {
        this.number = number;
    }
    
    
    /**
     * Add new label.
     */
    public void addLabel(String label) {
        this.labels.addElement(label);
    }
    
    
    /**
     * Remove a label.
     */
    public void removeLabel(String label) {
        this.labels.removeElement(label);
    }
    
    
    /**
     * Remove all labels.
     */
    public void removeLabels() {
        this.labels.removeAllElements();
    }
    
    
    /**
     * Remove specified labels.
     */
    /*    public void removeLabels(Collection c) {
     this.labels.removeAll(c);
     }*/
    
    
    /**
     * Enumerate labels.
     */
    public Enumeration enumerateLabels() {
        return this.labels.elements();
    }
    
    
    /**
     * Properties accessor.
     */
    Hashtable getProperties() {
        return this.properties;
    }
    
    
    /**
     * Properties mutator.
     */
    void setProperties(Hashtable properties) {
        // FIXME : Do a clean merge
        this.properties = properties;
    }
    
    
    /**
     * Tests if a property has been set.
     *
     * @param name Property name
     * @return true if the property has been set
     */
    public boolean exists(String name) {
        return exists(name, NodeProperty.DEFAULT_NAMESPACE);
    }
    
    
    /**
     * Tests if a property has been set.
     *
     * @param name Property name
     * @return true if the property has been set
     */
    public boolean exists(String name, String namespace) {
        if (name != null)
            return (properties.get(namespace + name) != null);
        else
            return (false);
    }
    
    
    /**
     * Property accessor.
     *
     * @param name Property name
     * @return String property value
     */
    public NodeProperty getProperty(String name) {
        return getProperty(name, NodeProperty.DEFAULT_NAMESPACE);
    }
    
    
    /**
     * Property accessor.
     *
     * @param name Property name
     * @param namespace Property namespace
     * @return String property value
     */
    public NodeProperty getProperty(String name, String namespace) {
        Object result = properties.get(namespace + name);
        if (result != null) {
            return (NodeProperty) result;
        } else {
            return null;
        }
    }
    
    
    /**
     * Property mutator.
     *
     * @param name Property name
     * @param value Property value
     */
    public void setProperty(String name, Object value) {
        setProperty(new NodeProperty(name, value));
    }
    
    
    /**
     * Property mutator.
     *
     * @param name Property name
     * @param value Property value
     * @param namespace Property value
     */
    public void setProperty(String name, String namespace, Object value) {
        setProperty(new NodeProperty(name, value, namespace));
    }
    
    
    /**
     * Property mutator.
     *
     * @param name Property name
     * @param value Property value
     * @param standard True if one of the standrad properties
     */
    protected void setProperty(String name, Object value, boolean standard) {
        setProperty(new NodeProperty(name, value, standard));
    }
    
    
    /**
     * Property mutatory.
     *
     * @param property Property
     */
    public void setProperty(NodeProperty property) {
        properties.put(property.getNamespace() + property.getName(),
                       property);
    }
    
    
    /**
     * Remove a property.
     *
     * @param property Property
     */
    public void removeProperty(NodeProperty property) {
        properties.remove(property.getNamespace() + property.getName());
    }
    
    
    /**
     * Remove a property.
     *
     * @param property Property
     */
    public void removeProperty(String property) {
        properties.remove(NodeProperty.DEFAULT_NAMESPACE + property);
    }
    
    
    /**
     * Remove a property.
     *
     * @param property Property
     */
    public void removeProperty(String property, String nameSpace) {
        properties.remove(nameSpace + property);
    }
    
    
    /**
     * Properties names enumerator.
     *
     * @return Enumeration of the properties names
     * @deprecated Replaced by enumeratePropertiesName
     */
    public Enumeration getPropertiesNames() {
        return enumeratePropertiesName();
    }
    
    
    /**
     * Properties names enumerator.
     *
     * @return Enumeration of the properties names
     */
    public Enumeration enumeratePropertiesName() {
        Vector result = new Vector();
        Enumeration propertyList = enumerateProperties();
        while (propertyList.hasMoreElements()) {
            NodeProperty currentProperty =
                (NodeProperty) propertyList.nextElement();
            result.addElement(currentProperty.getName());
        }
        return result.elements();
    }
    
    
    /**
     * Properties values enumerator.
     *
     * @return Enumeration of the properties values
     * @deprecated Replaced by enumerate properties
     */
    public Enumeration getPropertiesValues() {
        return enumerateProperties();
    }
    
    
    /**
     * Properties values enumerator.
     *
     * @return Enumeration of the properties values
     */
    public Enumeration enumerateProperties() {
        return properties.elements();
    }
    
    
    /**
     * Name accessor.
     *
     * @return String name
     */
    public String getName() {
        NodeProperty name = getProperty(NAME);
        if (name == null) {
            return new String();
        } else {
            return (String) name.getValue();
        }
    }
    
    
    /**
     * Name mutator.
     *
     * @param name New name
     */
    public void setName(String name) {
        setProperty(NAME, "<![CDATA[" + name + "]]>", true);
    }
    
    
    /**
     * Get the MIME content type of the data (if any).
     *
     * @return String
     */
    public String getContentType() {
        NodeProperty contentType = getProperty(CONTENT_TYPE);
        if (contentType == null) {
            return new String();
        } else {
            return (String) contentType.getValue();
        }
    }
    
    
    /**
     * Content type mutator.
     *
     * @param contentType New content type
     */
    public void setContentType(String contentType) {
        setProperty(CONTENT_TYPE, contentType, true);
    }
    
    
    /**
     * Content language accessor.
     *
     * @return String content language
     */
    public String getContentLanguage() {
        NodeProperty contentLanguage = getProperty(CONTENT_LANGUAGE);
        if (contentLanguage == null) {
            return new String();
        } else {
            return (String) contentLanguage.getValue();
        }
    }
    
    
    /**
     * Content language mutator.
     *
     * @param contentLanguage New content language
     */
    public void setContentLanguage(String contentLanguage) {
        setProperty(CONTENT_LANGUAGE, contentLanguage, true);
    }
    
    
    /**
     * Creation date accessor.
     *
     * @return String creation date
     */
    public String getCreationDate() {
        NodeProperty creationDate = getProperty(CREATION_DATE);
        if (creationDate == null) {
            return null;
        } else {
            if (creationDate.getValue() instanceof Date) {
                return creationDateFormat.format
                    ((Date) creationDate.getValue());
            }
            return creationDate.getValue().toString();
        }
    }
    
    
    /**
     * Creation date accessor.
     *
     * @return String creation date
     */
    public Date getCreationDateAsDate() {
        NodeProperty creationDate = getProperty(CREATION_DATE);
        if (creationDate == null)
            return null;
        if (creationDate.getValue() instanceof Date) {
            return (Date) creationDate.getValue();
        } else {
            String creationDateValue = creationDate.getValue().toString();
            Date result = null;
            // Parsing the HTTP Date
            for (int i = 0; (result == null) && (i < formats.length); i++) {
                try {
                    synchronized (formats[i]) {
                        result = formats[i].parse(creationDateValue);
                    }
                } catch (ParseException e) {
                    ;
                }
            }
            return result;
        }
    }
    
    
    /**
     * Creation date mutator.
     *
     * @param creationDate New creation date
     */
    public void setCreationDate(Date creationDate) {
        setProperty
            (CREATION_DATE, creationDateFormat.format(creationDate), true);
    }
    
    
    /**
     * Creation date mutator.
     *
     * @param creationDate New creation date
     */
    public void setCreationDate(String creationDate) {
        setProperty(CREATION_DATE, creationDate, true);
    }
    
    
    /**
     * Last modification date accessor.
     *
     * @return String last modification date
     */
    public String getLastModified() {
        NodeProperty lastModified = getProperty(LAST_MODIFIED);
        if (lastModified == null) {
            return null;
        } else {
            if (lastModified.getValue() instanceof Date) {
                return format.format((Date) lastModified.getValue());
            } else {
                return lastModified.getValue().toString();
            }
        }
    }
    
    
    /**
     * Creation date accessor.
     *
     * @return String creation date
     */
    public Date getLastModifiedAsDate() {
        NodeProperty lastModified = getProperty(LAST_MODIFIED);
        if (lastModified == null)
            return null;
        if (lastModified.getValue() instanceof Date) {
            return (Date) lastModified.getValue();
        } else {
            String lastModifiedValue = lastModified.getValue().toString();
            Date result = null;
            // Parsing the HTTP Date
            for (int i = 0; (result == null) && (i < formats.length); i++) {
                try {
                    synchronized (formats[i]) {
                        result = formats[i].parse(lastModifiedValue);
                    }
                } catch (ParseException e) {
                    ;
                }
            }
            return result;
        }
    }
    
    
    /**
     * Last modified mutator.
     *
     * @param lastModified New last modified date
     */
    public void setLastModified(Date lastModified) {
        setProperty(LAST_MODIFIED, format.format(lastModified), true);
    }
    
    
    /**
     * Last modified mutator.
     *
     * @param lastModified New last modified
     */
    public void setLastModified(String lastModified) {
        setProperty(LAST_MODIFIED, lastModified, true);
    }
    
    
    /**
     * Creation length mutator.
     *
     * @param creationLength New content length
     */
    public void setContentLength(long contentLength) {
        setProperty(CONTENT_LENGTH, new Long(contentLength), true);
    }
    
    
    /**
     * Creation length mutator.
     *
     * @param creationLength New content length
     */
    public void setContentLength(String contentLength) {
        Long contentLengthValue = null;
        try {
            contentLengthValue = new Long(contentLength);
        } catch (NumberFormatException e) {
            // Ignore
        }
        if (contentLengthValue == null) {
            contentLengthValue = new Long(0);
        }
        setProperty(CONTENT_LENGTH, contentLengthValue, true);
    }
    
    
    /**
     * Content length accessor.
     *
     * @return String
     */
    public long getContentLength() {
        NodeProperty contentLength = getProperty(CONTENT_LENGTH);
        if (contentLength == null) {
            return -1L;
        } else {
            if (contentLength.getValue() instanceof Long) {
                return ((Long) contentLength.getValue()).longValue();
            }
            
            if (contentLength.getValue() instanceof String) {
                return (new Long((String) contentLength.getValue()))
                    .longValue();
            }
            return -1L;
        }
    }
    
    
    /**
     * Set default properties.
     */
    void setDefaultProperties(Enumeration defaultProperties) {
        while (defaultProperties.hasMoreElements()) {
            NodeProperty currentProperty =
                (NodeProperty) defaultProperties.nextElement();
            String name = currentProperty.getName();
            String namespace = currentProperty.getNamespace();
            NodeProperty lookup = getProperty(name, namespace);
            if (lookup == null) {
                // Adding property
                setProperty(name, namespace, currentProperty.getValue());
            }
        }
    }
    
    
    // -------------------------------------------------------- Private methods
    
    
    private void initDefaultProperties() {
        
        setCreationDate(new Date());
        setName("");
        // By default, a resource is a collection
        setProperty(TYPE, "<collection/>", true);
        setProperty(SOURCE, "", true);
        setContentLength(-1);
        setLastModified(new Date());
        
    }
    
    
    // --------------------------------------------------------- Object Methods
    
    
    /**
     * Clone.
     *
     * @return Object clone
     */
    public NodeRevisionDescriptor cloneObject() {
        NodeRevisionDescriptor result = null;
        try {
            result = (NodeRevisionDescriptor) super.clone();
            // Cloning properties
            Hashtable propertiesClone = new Hashtable();
            Enumeration propertiesList = this.properties.keys();
            while (propertiesList.hasMoreElements()) {
                Object key = propertiesList.nextElement();
                Object value = this.properties.get(key);
                propertiesClone.put(key, value);
            }
            result.properties = propertiesClone;
            // Cloning labels
            result.labels = (Vector) this.labels.clone();
        } catch(CloneNotSupportedException e) {
            e.printStackTrace();
        }
        return result;
    }
    
    
    /**
     * Equals.
     *
     * @param obj Object to test
     * @return boolean True if the two object are equal :
     * <li>obj is of type NodeRevisionDescriptor and is not null</li>
     * <li>The two revision numbers are equal</li>
     */
    public boolean equals(Object obj) {
        boolean result = false;
        if ((obj != null) && (obj instanceof NodeRevisionDescriptor)) {
            NodeRevisionDescriptor revisionDescriptor =
                (NodeRevisionDescriptor) obj;
            result = this.getRevisionNumber()
                .equals(revisionDescriptor.getRevisionNumber());
        }
        return result;
    }
    
    
    /**
     * Validate.
     */
    public void validate() {
        
        if (branchName == null)
            throw new ObjectValidationFailedException
                (Messages.message
                     (NodeRevisionDescriptor.class.getName() + ".nullBranchName"));
        
        if (number == null)
            throw new ObjectValidationFailedException
                (Messages.message
                     (NodeRevisionDescriptor.class.getName() + ".nullNumber"));
        number.validate();
        
        if (labels == null)
            throw new ObjectValidationFailedException
                (Messages.message
                     (NodeRevisionDescriptor.class.getName() + ".nullLabels"));
        
        if (properties == null)
            throw new ObjectValidationFailedException
                (Messages.message
                     (NodeRevisionDescriptor.class.getName() + ".nullProperties"));
        Enumeration propertyList = properties.elements();
        while (propertyList.hasMoreElements()) {
            Object obj = propertyList.nextElement();
            if (!(obj instanceof NodeProperty))
                throw new ObjectValidationFailedException
                    (Messages.message
                         (NodeRevisionDescriptor.class.getName()
                              + ".invalidPropertyType"));
            ((NodeProperty) obj).validate();
        }
        
    }
    
    
}
