/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/util/Messages.java,v 1.4 2000/09/26 02:44:29 remm Exp $
 * $Revision: 1.4 $
 * $Date: 2000/09/26 02:44:29 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:  
 *       "This product includes software developed by the 
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */ 

package org.apache.slide.util;

import java.text.*;
import java.util.*;
import org.apache.slide.common.Domain;

/**
 * Messages resource bundle manager.
 * 
 * @author <a href="mailto:arkin@intalio.com">Assaf Arkin</a>
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.4 $ $Date: 2000/09/26 02:44:29 $
 */
public class Messages {
    
    
    // -------------------------------------------------------------- Constants
    
    
    /**
     * Resource filename.
     */
    public static final String ResourceName = 
        "org.apache.slide.util.resources.messages";
    
    
    // -------------------------------------------------------------- Variables
    
    
    /**
     * Resource bundle.
     */
    private static ResourceBundle _messages;
    
    
    /**
     * Formats.
     */
    private static Hashtable _formats;
    
    
    // --------------------------------------------------------- Public Methods
    
    
    /**
     * Format a message.
     * 
     * @param message Message to format
     * @param arg1 Argument 1
     */
    public static String format(String message, Object arg1) {
        return format( message, new Object[] { arg1 } );
    }
    
    
    /**
     * Format a message.
     * 
     * @param message Message to format
     * @param arg1 Argument 1
     * @param arg2 Argument 2
     */
    public static String format(String message, Object arg1, Object arg2) {
        return format( message, new Object[] { arg1, arg2 } );
    }
    
    
    /**
     * Format a message.
     * 
     * @param message Message to format
     * @param arg1 Argument 1
     * @param arg2 Argument 2
     * @param arg3 Argument 3
     */
    public static String format(String message, Object arg1, Object arg2, 
                                Object arg3) {
        return format( message, new Object[] { arg1, arg2, arg3 } );
    }
    
    
    /**
     * Format a message.
     * 
     * @param message Message to format
     * @param arg1 Argument array
     */
    public static String format( String message, Object[] args )
    {
        MessageFormat mf;
        String msg;
        
        try {
            mf = (MessageFormat) _formats.get( message );
            if (mf == null) {
                try {
                    msg = _messages.getString( message );
                } catch (MissingResourceException except) {
                    return message;
                }
                mf = new MessageFormat(msg);
                _formats.put(message, mf);
            }
            return mf.format(args);
        } catch (Exception except) {
            return "An internal error occured while processing message " 
                + message;
        }
    }
    
    
    /**
     * Fetch a message from the resource bundle.
     * 
     * @param message Id of the message if the bundle
     */
    public static String message(String message) {
        try {
            return _messages.getString(message);
        } catch (MissingResourceException except) {
            return message;
        }
    }
    
    
    /**
     * Set current locale.
     * 
     * @param locale Current locale
     */
    public static void setLocale(Locale locale) {
        _formats = new Hashtable();
        try {
            if (locale == null) {
                _messages = ResourceBundle.getBundle(ResourceName); 
            } else {
                _messages = ResourceBundle.getBundle(ResourceName, locale);
            }
        } catch (Exception except) {
            _messages = new EmptyResourceBundle();
            Domain.error("Failed to locate messages resource " + ResourceName);
        }
    }
    
    
    // ------------------------------------------------------------ Initializer
    
    
    static {
        setLocale(Locale.getDefault());
    }
    
    
    // ---------------------------------------- EmptyResourceBundle Inner Class
    
    
    static class EmptyResourceBundle
        extends ResourceBundle
        implements Enumeration {
        
        public Enumeration getKeys() {
            return this;
        }
        
        protected Object handleGetObject(String name) {
            return "[Missing message " + name + "]";
        }
        
        public boolean hasMoreElements() {
            return false;
        }
        
        public Object nextElement() {
            return null;
        }
        
    }
    
}
