/*
 * $Header: /home/cvs/jakarta-slide/src/taglib/struts/org/apache/slide/taglib/tag/struts/IterateMembersTag.java,v 1.2 2001/09/21 16:59:02 cmlenz Exp $
 * $Revision: 1.2 $
 * $Date: 2001/09/21 16:59:02 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:  
 *       "This product includes software developed by the 
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */ 

package org.apache.slide.taglib.tag.struts;

import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.PageContext;

import org.apache.slide.common.SlideException;
import org.apache.slide.taglib.bean.NodeBean;
import org.apache.struts.taglib.logic.IterateTag;


/**
 * Tag class that extends the generic Struts iteration tag to provide iteration
 * over the members of a specific node up to a configurable depth, including 
 * and/or excluding specific types and/or roles of objects.
 * 
 * @author <a href="mailto:cmlenz@apache.org">Christopher Lenz</a>
 * @version $Revision: 1.2 $
 */
public class IterateMembersTag
    extends IterateTag {
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Value of the 'depth' attribute.
     */
    protected String depth;
    
    
    /**
     * Value of the 'includeRoles' attribute.
     */
    protected String includeRoles;
    
    
    /**
     * Value of the 'excludeRoles' attribute.
     */
    protected String excludeRoles;
    
    
    /**
     * Value of the 'includeTypes' attribute.
     */
    protected String includeTypes;
    
    
    /**
     * Value of the 'excludeTypes' attribute.
     */
    protected String excludeTypes;
    
    
    // ------------------------------------------------------------ Tag Methods
    
    
    /**
     * Called by the JSP Engine when opening this tag.
     *
     * @throws JspException not thrown
     */
    public int doStartTag()
        throws JspException {
        
        NodeBean node = StrutsTagUtils.findNode(this, pageContext);
        if (node == null) {
            throw new JspException("The 'iterateMembers' tag must be nested " +
                "inside a 'node' tag or an iteration over nodes.");
        }
        
        int depthValue = 1;
        if (depth != null) {
            try {
                depthValue = Integer.parseInt(depth);
            } catch (NumberFormatException e) {
                if (depth.equalsIgnoreCase("infinity")) {
                    depthValue = Integer.MAX_VALUE;
                }
                else {
                    throw new JspException("The attribute 'depth' of the " +
                        "'iterateMembers' tag must contain an integer value.");
                }
            }
        }
        
        Vector members = node.getMembers(depthValue);
        
        // evaluate the includes/excludes, which can be comma separated lists
        Vector includeRolesVector = tokenizeString(includeRoles);
        Vector excludeRolesVector = tokenizeString(excludeRoles);
        Vector includeTypesVector = tokenizeString(includeTypes);
        Vector excludeTypesVector = tokenizeString(excludeTypes);
        
        Iterator i = members.iterator();
        membersIteration: while (i.hasNext()) {
            NodeBean member = (NodeBean)i.next();
            
            String nodeType = member.getType();
            if (excludeTypesVector.contains(nodeType) ||
                (!includeTypesVector.isEmpty() && 
                 !includeTypesVector.contains(nodeType))) {
                i.remove();
                continue membersIteration;
            }
            
            Vector nodeRoles = member.getRoles();
            for (int j = 0; j < nodeRoles.size(); j++) {
                String nodeRole = (String)nodeRoles.elementAt(j);
                if (excludeRolesVector.contains(nodeRole)) {
                    i.remove();
                    continue membersIteration;
                }
                if (!includeRolesVector.isEmpty() && 
                    includeRolesVector.contains(nodeRole)) {
                    continue membersIteration;
                }
            }
            
            // if this point is reached, none of the member's roles has been 
            // included, so we remove it from the list
            if (!includeRolesVector.isEmpty()) {
                i.remove();
                continue membersIteration;
            }
        }
        
        setCollection(members);
        
        return super.doStartTag();
    }
    
    
    /**
     * Releases any resources we may have (or inherit).
     */
    public void release() {
        super.release();
        
        depth = null;
        includeRoles = null;
        excludeRoles = null;
        includeTypes = null;
        excludeTypes = null;
    }
    
    
    // ------------------------------------------------------------- Properties
    
    
    /**
     * Returns the 'depth' attribute.
     *
     * @return value of the 'depth' attribute
     */
    public String getDepth() {
        
        return depth;
    }
    
    
    /**
     * Sets the 'depth' attribute.
     *
     * @param depth the attribute value
     */
    public void setDepth(String depth) {
        
        this.depth = depth;
    }
    
    
    /**
     * Sets the 'excludeTypes' attribute.
     *
     * @param excludeTypes the attribute value
     */
    public void setExcludeTypes(String excludeTypes) {
        
        this.excludeTypes = excludeTypes;
    }
    
    
    /**
     * Sets the 'includeTypes' attribute.
     *
     * @param includeTypes the attribute value
     */
    public void setIncludeTypes(String includeTypes) {
        
        this.includeTypes = includeTypes;
    }
    
    
    /**
     * Sets the 'excludeRoles' attribute.
     *
     * @param excludeRoles the attribute value
     */
    public void setExcludeRoles(String excludeRoles) {
        
        this.excludeRoles = excludeRoles;
    }
    
    
    /**
     * Sets the 'includeRoles' attribute.
     *
     * @param includeRoles the attribute value
     */
    public void setIncludeRoles(String includeRoles) {
        
        this.includeRoles = includeRoles;
    }
    
    
    // -------------------------------------------------------- Private Methods
    
    
    /**
     * Splits the given string into its tokens, separated by commata, and 
     * returns the tokens in a new Vector.
     */
    private static Vector tokenizeString(String values) {
        
        Vector v = new Vector();
        if (values != null) {
            StringTokenizer st = new StringTokenizer(values, ",");
            while (st.hasMoreElements()) {
                v.addElement(((String)st.nextElement()).trim());
            }
        }
        
        return v;
    }
    
    
}

