/*
 * $Header: /home/cvs/jakarta-slide/src/webdav/client/src/org/apache/webdav/lib/WebdavResources.java,v 1.2 2001/05/16 16:51:41 jericho Exp $
 * $Revision: 1.2 $
 * $Date: 2001/05/16 16:51:41 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */
package org.apache.webdav.lib;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This WebdavResources class represents WebDAV resources in a collection.
 *
 * @author <a href="mailto:jericho@thinkfree.com">Park, Sung-Gu</a>
 */

public class WebdavResources {

    // --------------------------------------------------------- Constructors

    /**
     * Default constuctor.
     */
    public WebdavResources() {
    }


    /**
     * This constuctor.
     *
     * @param resource A resource to add.
     */
    public WebdavResources(WebdavResource resource) {
        addResource(resource);
    }


    // --------------------------------------------------- Instance Variables


    /**
     * The resources for each href and its properties.
     */
    Hashtable hrefTable = new Hashtable();


    // ------------------------------------------------------- Public Methods


    /**
     * Test if there is no resource.
     *
     * @return true if it's empty.
     */
    public boolean isEmpty() {
        return hrefTable.isEmpty();
    }


    /**
     * Test if there is a resource called the specified resource name.
     *
     * @param resourceName The resource name to check.
     */
    public boolean isThereResourceName(String resourceName) {
        return hrefTable.containsKey(resourceName) ? true : false;
    }


    /**
     * Test if there is a resource.
     *
     * @param resource The specified resource. 
     * @return true if it exists.
     */
    public boolean isThereResource(WebdavResource resource) {
        return hrefTable.contains(resource) ? true : false;
    }


    /**
     * Get an enumeration of the resource names.
     *
     * @return An enumeration of the resource names.
     */ 
    public Enumeration getResourceNames() {
        return hrefTable.keys();
    }


    /**
     * Get an enumeration of the resources.
     *
     * @return An enumeration of resources.
     */
    public Enumeration getResources() {
        return hrefTable.elements();
    }


    /**
     * Get an array of resource names.
     *
     * @return An array of resource names.
     */
    public String[] list() {
        
        synchronized (hrefTable) { 
            int num = hrefTable.size();
            String resourceNames[] = new String[num];

            Enumeration resources = getResourceNames();
            int i = 0;
            while (resources.hasMoreElements()) {
                resourceNames[i++] = (String) resources.nextElement();
            }

            return resourceNames;
        }
    }


    /**
     * Get an arraay of resources.
     * 
     * @return An array of resources.
     */
    public WebdavResource[] listResources() {
        
        synchronized (hrefTable) { 
            int num = hrefTable.size();
            WebdavResource WebdavResources[] = new WebdavResource[num];

            Enumeration resources = getResources();
            int i = 0;
            while (resources.hasMoreElements()) {
                WebdavResources[i++] =
                    (WebdavResource) resources.nextElement();
            }

            return WebdavResources;
        }
    }


    /**
     * Get an resource.
     *
     * @param resourceName The resource name.
     * @return The wanted resource if it exists.
     */
    public WebdavResource getResource(String resourceName) {
        return (WebdavResource) hrefTable.get(resourceName);
    }


    /**
     * Maps the resource name to its resource.
     * The resource name could be different from the displayname property.
     * It's useful for representing itself or parent collection.
     *
     * @param resourceName The resource name.
     * @param resource The resource.
     * @see #addResource(WebdavResource)
     */
    public void addResource(String resourceName, WebdavResource resource) {
        hrefTable.put(resourceName, resource);
    }


    /**
     * Add the specified resource.
     *
     * @param resource The resource to add.
     */
    public void addResource(WebdavResource resource) {
        hrefTable.put(resource.getDisplayName(), resource);
    }


    /**
     * Remove the specified resource name.
     *
     * @param resourceName The specified resource name.
     * @return The wanted resource.
     */
    public WebdavResource removeResource(String resourceName) {
        return (WebdavResource) hrefTable.remove(resourceName);
    }


    /**
     * Remove all resources.
     */
    public void removeAll() {
        hrefTable.clear();
    }


    /**
     * Return the string for this class.
     */
    public String toString() {
        return hrefTable.toString();
    }

}






