/*
 * $Header: /home/cvs/jakarta-slide/src/webdav/client/src/org/apache/webdav/lib/properties/GetLastModifiedProperty.java,v 1.4 2001/09/04 17:37:57 juergen Exp $
 * $Revision: 1.4 $
 * $Date: 2001/09/04 17:37:57 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */
package org.apache.webdav.lib.properties;

import java.util.Date;
import java.util.Locale;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;

import org.w3c.dom.Element;

import org.apache.util.DOMUtils;

import org.apache.webdav.lib.ResponseEntity;
import org.apache.webdav.lib.BaseProperty;

/**
 * This interface models the <code>&lt;D:getlastmodified&gt;</code> property,
 * which indicates the last time the resource was modified.  It does not, as
 * the name might misleadingly imply, have anything to do with getting the most
 * recently modified property.
 *
 * @author BC Holmes
 * @version $Revision: 1.4 $
 */
public class GetLastModifiedProperty extends BaseProperty {
    
    
    // -------------------------------------------------------------- Constants
    
    
    /**
     * The property name.
     */
    public static final String TAG_NAME = "getlastmodified";
    
    
    /**
     * The standard date format for the last modified date, as specified in
     * the HTTP 1.1 specification (RFC 2068).
     */
    public static final String DATE_FORMAT = "EEE, d MMM yyyy kk:mm:ss z";
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Default constructor for the property.
     */
    public GetLastModifiedProperty(ResponseEntity response, Element element) {
        super(response, element);
        // RFC 1123, 822. Date and time specification is English.
        format = new SimpleDateFormat(DATE_FORMAT, Locale.US);
    }
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    protected DateFormat format = null;
    
    
    // --------------------------------------------------------- Public Methods
    
    
    /**
     * Get the date of the last modification.
     *
     * @returns the last modified date.
     */
    public Date getDate() {
        Date date = null;
        try {
            synchronized (format) {
                date = format.parse(DOMUtils.getTextValue(element));
            }
        } catch (ParseException e) {
        }
        return date;
    }
    
    
}
