/*
 * $Header: /home/cvs/jakarta-slide/src/webdav/server/org/apache/slide/webdav/method/UnlockMethod.java,v 1.14 2001/09/10 07:16:17 juergen Exp $
 * $Revision: 1.14 $
 * $Date: 2001/09/10 07:16:17 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.slide.webdav.method;

import java.security.Principal;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.apache.util.WebdavStatus;
import org.apache.slide.common.*;
import org.apache.slide.webdav.*;
import org.apache.slide.macro.*;
import org.apache.slide.lock.*;
import org.apache.slide.content.*;
import org.apache.slide.security.AccessDeniedException;
import org.apache.slide.structure.*;

/**
 * UNLOCK method.
 *
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 */
public class UnlockMethod extends WebdavMethod {
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Resource to unlock.
     */
    private String toUnlock;
    
    
    /**
     * Id.
     */
    private String lockId;
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * UNLOCK method constructor.
     *
     * @param token Namespace access token
     * @param req HTTP request
     * @param resp HTTP response
     */
    public UnlockMethod(NamespaceAccessToken token, HttpServletRequest req,
                        HttpServletResponse resp, WebdavServletConfig config) {
        super(token, req, resp, config);
    }
    
    
    // ------------------------------------------------------ Protected Methods
    
    
    /**
     * Parse the request.
     *
     * @exception WebdavException Bad request
     */
    protected void parseRequest()
        throws WebdavException {
        
        // Loads the associated object from the store.
        toUnlock = requestUri;
        if (toUnlock == null) {
            toUnlock = "/";
        }
        
        String lockTokenHeader = req.getHeader("Lock-Token");
        if (lockTokenHeader != null) {
            lockId = parseLockToken(lockTokenHeader);
        } else {
            // TODO : The call to UNLOCK is invalid
            System.out.println("Invalid call");
        }
        
    }
    
    
    /**
     * Execute the request.
     *
     * @exception WebdavException
     */
    protected void executeRequest()
        throws WebdavException {
        
        // Prevent dirty reads
        slideToken.setForceStoreEnlistment(true);
        
        
        if (lockId != null) {
            
            try {
                
                lock.unlock(slideToken, requestUri, lockId);
                
                // Checking if the resource at the URI isn't a lock-null
                // resource, in which case we must attempt to delete it
                ObjectNode node = structure.retrieve(slideToken, requestUri);
                NodeRevisionDescriptors revisionDescriptors =
                    content.retrieve(slideToken, requestUri);
                NodeRevisionDescriptor revisionDescriptor =
                    content.retrieve(slideToken, revisionDescriptors);
                NodeProperty typeProperty =
                    revisionDescriptor.getProperty
                    (NodeRevisionDescriptor.TYPE);
                
                if (typeProperty.getValue().toString()
                    .equals("<lock-null/>")) {
                    
                    content.remove(slideToken, requestUri, revisionDescriptor);
                    content.remove(slideToken, revisionDescriptors);
                    structure.remove(slideToken, node);
                    
                }
                
                resp.setStatus(WebdavStatus.SC_NO_CONTENT);
                
        } catch (Exception e) {
            resp.setStatus(getErrorCode(e));  // special handling needed
            throw new WebdavException(WebdavStatus.SC_ACCEPTED, false); // abort the TA
        }
            
        }
        
    }
    
    
        
        
        
    /**
     * Get return status based on exception type.
     */
    protected int getErrorCode(Exception ex) {
        try {
            throw ex;
        } catch (RevisionDescriptorNotFoundException e) {
            return WebdavStatus.SC_OK;
        } catch (LinkedObjectNotFoundException e) {
            return WebdavStatus.SC_NOT_FOUND;
        } catch (Exception e) {
            return super.getErrorCode(e);
        }
    }


        
    
    /**
     * Returns true
     */
    protected boolean methodNeedsTransactionSupport() {
        return true;
    }
    
    
}
